@@ -35,7 +35,6 @@
 #include <string.h>
 #include <stdbool.h>
 #include <errno.h>
-#include <limits.h>
 
 #include <netinet/in.h>
 #include <netinet/sctp.h>
@@ -53,36 +52,17 @@
 
 #include "osmo_io_internal.h"
 
-#define IOFD_URING_INITIAL_SIZE 4096
-/* 32768 refers to the IORING_MAX_ENTRIES of the kernel (io_uring/io_uring.h). */
-#define IOFD_URING_MAXIMUM_SIZE 32768
-
-#define OSMO_IO_URING_BATCH "LIBOSMO_IO_URING_BATCH"
-
-#define OSMO_IO_URING_INITIAL_SIZE "LIBOSMO_IO_URING_INITIAL_SIZE"
-
-#define OSMO_IO_URING_READ_SQE "LIBOSMO_IO_URING_READ_SQE"
-
-bool g_io_uring_batch = false;
-bool g_io_uring_submit_needed = false;
-
-static int g_io_uring_size = IOFD_URING_INITIAL_SIZE;
-
-static int g_io_uring_read_sqes = 1;
+#define IOFD_URING_ENTRIES 4096
 
 struct osmo_io_uring {
 	struct osmo_fd event_ofd;
 	struct io_uring ring;
-	struct llist_head cancel_queue;
-	unsigned int num_pending_submissions;
 };
 
-static __thread struct osmo_io_uring *g_ring = NULL;
+static __thread struct osmo_io_uring g_ring;
 
 static void iofd_uring_cqe(struct io_uring *ring);
 
-void osmo_io_uring_submit(void);
-
 /*! read call-back for eventfd notifying us if entries are in the completion queue */
 static int iofd_uring_poll_cb(struct osmo_fd *ofd, unsigned int what)
 {
@@ -108,151 +88,43 @@ static int iofd_uring_poll_cb(struct osmo_fd *ofd, unsigned int what)
 /*! initialize the uring and tie it into our event loop */
 void osmo_iofd_uring_init(void)
 {
-	const char *env;
 	int rc, evfd;
 
-	if ((env = getenv(OSMO_IO_URING_BATCH)))
-		g_io_uring_batch = true;
-
-	if (!g_ring && (env = getenv(OSMO_IO_URING_INITIAL_SIZE))) {
-		int env_value;
-		rc = osmo_str_to_int(&env_value, env, 10, 1, IOFD_URING_MAXIMUM_SIZE);
-		if (rc < 0) {
-			fprintf(stderr, "Error: Initial io_uring size out of range (1..%d).\n",
-				IOFD_URING_MAXIMUM_SIZE);
-			exit(1);
-		}
-		if ((env_value & (env_value - 1))) {
-			fprintf(stderr, "Error: Initial io_uring size must be a positive power of two.\n");
-			exit(1);
-		}
-		g_io_uring_size = env_value;
-	}
-
-	g_ring = talloc_zero(OTC_GLOBAL, struct osmo_io_uring);
-	INIT_LLIST_HEAD(&g_ring->cancel_queue);
-
-	rc = io_uring_queue_init(g_io_uring_size, &g_ring->ring, 0);
+	rc = io_uring_queue_init(IOFD_URING_ENTRIES, &g_ring.ring, 0);
 	if (rc < 0)
 		osmo_panic("failure during io_uring_queue_init(): %s\n", strerror(-rc));
 
-	if ((env = getenv(OSMO_IO_URING_READ_SQE))) {
-		g_io_uring_read_sqes = atoi(env);
-		if (g_io_uring_read_sqes < 1 || g_io_uring_read_sqes > IOFD_MSGHDR_MAX_READ_SQES) {
-			fprintf(stderr, "Invalid osmo_uring read SQEs requested: \"%s\"\n Allowed range: 1..%d\n",
-				env, IOFD_MSGHDR_MAX_READ_SQES);
-			exit(1);
-		}
-	}
-
 	rc = eventfd(0, 0);
 	if (rc < 0) {
-		io_uring_queue_exit(&g_ring->ring);
+		io_uring_queue_exit(&g_ring.ring);
 		osmo_panic("failure creating eventfd(0, 0) for io_uring: %s\n", strerror(-rc));
 	}
 	evfd = rc;
 
-	osmo_fd_setup(&g_ring->event_ofd, evfd, OSMO_FD_READ, iofd_uring_poll_cb, &g_ring->ring, 0);
-	rc = osmo_fd_register(&g_ring->event_ofd);
+	osmo_fd_setup(&g_ring.event_ofd, evfd, OSMO_FD_READ, iofd_uring_poll_cb, &g_ring.ring, 0);
+	rc = osmo_fd_register(&g_ring.event_ofd);
 	if (rc < 0) {
 		close(evfd);
-		io_uring_queue_exit(&g_ring->ring);
+		io_uring_queue_exit(&g_ring.ring);
 		osmo_panic("failure registering io_uring-eventfd as osmo_fd: %d\n", rc);
 	}
-	rc = io_uring_register_eventfd(&g_ring->ring, evfd);
+	rc = io_uring_register_eventfd(&g_ring.ring, evfd);
 	if (rc < 0) {
-		osmo_fd_unregister(&g_ring->event_ofd);
+		osmo_fd_unregister(&g_ring.event_ofd);
 		close(evfd);
-		io_uring_queue_exit(&g_ring->ring);
+		io_uring_queue_exit(&g_ring.ring);
 		osmo_panic("failure registering eventfd with io_uring: %s\n", strerror(-rc));
 	}
 }
 
-static void osmo_iofd_uring_exit(struct osmo_io_uring *ring)
-{
-	LOGP(DLIO, LOGL_DEBUG, "Old empty io_uring will be destroyed.");
 
-	io_uring_queue_exit(&ring->ring);
-
-	osmo_fd_unregister(&ring->event_ofd);
-	close(ring->event_ofd.fd);
-
-	talloc_free(ring);
-}
-
-static inline struct io_uring_sqe *io_uring_get_sqe_and_count(struct osmo_io_uring *ring)
-{
-	struct io_uring_sqe *sqe;
-
-	sqe = io_uring_get_sqe(&ring->ring);
-	if (sqe)
-		ring->num_pending_submissions++;
-
-	return sqe;
-}
-
-static struct io_uring_sqe *iofd_uring_get_sqe(struct osmo_io_fd *iofd, bool read)
-{
-	struct io_uring_sqe *sqe;
-
-	/* All subsequent read SQEs must be on the same ring. */
-	if (read && iofd->u.uring.read.sqes_submitted > 0 && iofd->u.uring.read.ring != &g_ring->ring)
-		return NULL;
-
-	sqe = io_uring_get_sqe_and_count(g_ring);
-	if (sqe)
-		return sqe;
-
-	/* The current ring is full, subsequent read SQEs on different ring are not allowed. */
-	if (read && iofd->u.uring.read.sqes_submitted > 0)
-		return NULL;
-
-	if (g_io_uring_size < IOFD_URING_MAXIMUM_SIZE) {
-		LOGP(DLIO, LOGL_NOTICE, "io_uring too small to handle all SQEs with its current size of %d. "
-		     "Increasing io_uring size to %d.\n", g_io_uring_size, g_io_uring_size * 2);
-		g_io_uring_size <<= 1;
-	} else {
-		LOGP(DLIO, LOGL_NOTICE, "io_uring too small to handle all SQEs with its maximum size of %d. "
-		     "adding another one.\n", g_io_uring_size);
-	}
-
-	/* Submit all SQEs of current ring and create a new ring, if needed.
-	 * The old ring will be kept, as there are uncompleted submissions.
-	 * TODO: Destroy old ring, once all submissions are completed. */
-	osmo_io_uring_submit();
-
-	osmo_iofd_uring_init();
-
-	sqe = io_uring_get_sqe_and_count(g_ring);
-	OSMO_ASSERT(sqe);
-	return sqe;
-}
-
-static inline void iofd_io_uring_submit(void)
-{
-	if (OSMO_LIKELY(!g_io_uring_batch))
-		io_uring_submit(&g_ring->ring);
-	else
-		g_io_uring_submit_needed = true;
-}
-
-static inline int iofd_uring_submit_recv_sqe(struct osmo_io_fd *iofd, enum iofd_msg_action action)
+static void iofd_uring_submit_recv(struct osmo_io_fd *iofd, enum iofd_msg_action action)
 {
 	struct msgb *msg;
 	struct iofd_msghdr *msghdr;
 	struct io_uring_sqe *sqe;
-	uint8_t idx;
 
-	/* Tell iofd_uring_get_sqe() not to allocate a new ring, if we want to enqueue multiple read SQEs. */
-	sqe = iofd_uring_get_sqe(iofd, true);
-	if (!sqe) {
-		if (iofd->u.uring.read.sqes_submitted > 0)
-			return -EINVAL;
-		LOGPIO(iofd, LOGL_ERROR, "Could not get io_uring_sqe\n");
-		OSMO_ASSERT(0);
-	}
-
-	msg = iofd_msgb_alloc(iofd);
+	msg = iofd_msgb_pending_or_alloc(iofd);
 	if (!msg) {
 		LOGPIO(iofd, LOGL_ERROR, "Could not allocate msgb for reading\n");
 		OSMO_ASSERT(0);
@@ -264,10 +136,8 @@ static inline int iofd_uring_submit_recv_sqe(struct osmo_io_fd *iofd, enum iofd_
 		OSMO_ASSERT(0);
 	}
 
-	for (idx = 0; idx < msghdr->io_len; idx++) {
-		msghdr->iov[idx].iov_base = msghdr->msg[idx]->tail;
-		msghdr->iov[idx].iov_len = msgb_tailroom(msghdr->msg[idx]);
-	}
+	msghdr->iov[0].iov_base = msg->tail;
+	msghdr->iov[0].iov_len = msgb_tailroom(msg);
 
 	switch (action) {
 	case IOFD_ACT_RECVMSG:
@@ -280,15 +150,21 @@ static inline int iofd_uring_submit_recv_sqe(struct osmo_io_fd *iofd, enum iofd_
 		/* fall-through */
 	case IOFD_ACT_READ:
 		msghdr->hdr.msg_iov = &msghdr->iov[0];
-		msghdr->hdr.msg_iovlen = msghdr->io_len;
+		msghdr->hdr.msg_iovlen = 1;
 		break;
 	default:
 		OSMO_ASSERT(0);
 	}
 
+	sqe = io_uring_get_sqe(&g_ring.ring);
+	if (!sqe) {
+		LOGPIO(iofd, LOGL_ERROR, "Could not get io_uring_sqe\n");
+		OSMO_ASSERT(0);
+	}
+
 	switch (action) {
 	case IOFD_ACT_READ:
-		io_uring_prep_readv(sqe, iofd->fd, msghdr->iov, msghdr->hdr.msg_iovlen, -1);
+		io_uring_prep_readv(sqe, iofd->fd, msghdr->iov, 1, -1);
 		break;
 	case IOFD_ACT_RECVMSG:
 	case IOFD_ACT_RECVFROM:
@@ -299,81 +175,28 @@ static inline int iofd_uring_submit_recv_sqe(struct osmo_io_fd *iofd, enum iofd_
 	}
 	io_uring_sqe_set_data(sqe, msghdr);
 
-	iofd_io_uring_submit();
-
-	iofd->u.uring.read.msghdr[iofd->u.uring.read.sqes_submitted] = msghdr;
-	iofd->u.uring.read.sqes_submitted++;
-	iofd->u.uring.read.ring = &g_ring->ring;
-
-	return 0;
-}
-
-static void iofd_uring_submit_recv(struct osmo_io_fd *iofd, enum iofd_msg_action action)
-{
-	int rc;
-
-	/* Submit more read SQEs in advance, if requested. */
-	while (iofd->u.uring.read.sqes_submitted < iofd->u.uring.read.num_sqes) {
-		rc = iofd_uring_submit_recv_sqe(iofd, action);
-		/* Stop, if we cannot enqueue multiple read SQEs in the same ring. */
-		if (rc < 0)
-			break;
-	}
+	io_uring_submit(&g_ring.ring);
+	/* NOTE: This only works if we have one read per fd */
+	iofd->u.uring.read_msghdr = msghdr;
 }
 
 /*! completion call-back for READ/RECVFROM */
 static void iofd_uring_handle_recv(struct iofd_msghdr *msghdr, int rc)
 {
 	struct osmo_io_fd *iofd = msghdr->iofd;
-	uint8_t idx, i;
+	struct msgb *msg = msghdr->msg;
 
-	/* Find which read_msghdr is completed and remove from list. */
-	for (idx = 0; idx < iofd->u.uring.read.sqes_submitted; idx++) {
-		if (iofd->u.uring.read.msghdr[idx] == msghdr)
-			break;
-	}
-	if (idx == iofd->u.uring.read.sqes_submitted) {
-		LOGP(DLIO, LOGL_FATAL, "Read SQE completion, but msghdr not found, please fix!\n");
-		return;
-	}
-	/* Remove entry at idx. */
-	iofd->u.uring.read.sqes_submitted--;
-	for (i = idx; i < iofd->u.uring.read.sqes_submitted; i++)
-		iofd->u.uring.read.msghdr[i] = iofd->u.uring.read.msghdr[i + 1];
-	iofd->u.uring.read.msghdr[i] = NULL;
+	if (rc > 0)
+		msgb_put(msg, rc);
 
-	for (idx = 0; idx < msghdr->io_len; idx++) {
-		struct msgb *msg = msghdr->msg[idx];
-		int chunk;
+	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED))
+		iofd_handle_recv(iofd, msg, rc, msghdr);
 
-		msghdr->msg[idx] = NULL;
-		if (rc > 0) {
-			if (rc > msghdr->iov[idx].iov_len)
-				chunk = msghdr->iov[idx].iov_len;
-			else
-				chunk = rc;
-			rc -= chunk;
-			msgb_put(msg, chunk);
-		} else {
-			chunk = rc;
-		}
-
-		/* Check for every iteration, because iofd might get unregistered/closed during receive function. */
-		if (iofd->u.uring.read.enabled && !IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))
-			iofd_handle_recv(iofd, msg, chunk, msghdr);
-		else
-			msgb_free(msg);
-
-		if (rc <= 0)
-			break;
-	}
-	while (++idx < msghdr->io_len) {
-		msgb_free(msghdr->msg[idx]);
-		msghdr->msg[idx] = NULL;
-	}
-
-	if (iofd->u.uring.read.enabled && !IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))
+	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED) && iofd->u.uring.read_enabled)
 		iofd_uring_submit_recv(iofd, msghdr->action);
+	else
+		iofd->u.uring.read_msghdr = NULL;
+
 
 	iofd_msghdr_free(msghdr);
 }
@@ -388,26 +211,21 @@ static void iofd_uring_handle_tx(struct iofd_msghdr *msghdr, int rc)
 	/* Detach msghdr from iofd. It might get freed here or it is freed during iofd_handle_send_completion().
 	 * If there is pending data to send, iofd_uring_submit_tx() will attach it again.
 	 * iofd_handle_send_completion() will invoke a callback function to signal the possibility of write/send.
-	 * This callback function might close iofd, leading to the potential freeing of iofd->u.uring.write.msghdr if
+	 * This callback function might close iofd, leading to the potential freeing of iofd->u.uring.write_msghdr if
 	 * still attached. Since iofd_handle_send_completion() frees msghdr at the end of the function, detaching
 	 * msghdr here prevents a double-free bug. */
-	if (iofd->u.uring.write.msghdr == msghdr) {
-		iofd->u.uring.write.msghdr = NULL;
-		iofd->u.uring.write.ring = NULL;
-	}
+	if (iofd->u.uring.write_msghdr == msghdr)
+		iofd->u.uring.write_msghdr = NULL;
 
 	if (OSMO_UNLIKELY(IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))) {
-		for (int idx = 0; idx < msghdr->io_len; idx++) {
-			msgb_free(msghdr->msg[idx]);
-			msghdr->msg[idx] = NULL;
-		}
+		msgb_free(msghdr->msg);
 		iofd_msghdr_free(msghdr);
 	} else {
 		iofd_handle_send_completion(iofd, rc, msghdr);
 	}
 
 	/* submit the next to-be-transmitted message for this file descriptor */
-	if (iofd->u.uring.write.enabled && !IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))
+	if (iofd->u.uring.write_enabled && !IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))
 		iofd_uring_submit_tx(iofd);
 }
 
@@ -435,7 +253,7 @@ static void iofd_uring_handle_completion(struct iofd_msghdr *msghdr, int res)
 
 	IOFD_FLAG_UNSET(iofd, IOFD_FLAG_IN_CALLBACK);
 
-	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_TO_FREE) && !iofd->u.uring.read.sqes_submitted && !iofd->u.uring.write.msghdr)
+	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_TO_FREE) && !iofd->u.uring.read_msghdr && !iofd->u.uring.write_msghdr)
 		talloc_free(iofd);
 }
 
@@ -444,11 +262,9 @@ static void iofd_uring_cqe(struct io_uring *ring)
 {
 	int rc;
 	struct io_uring_cqe *cqe;
-	struct iofd_msghdr *msghdr, *msghdr2;
-	struct osmo_io_uring *orig_ring = container_of(ring, struct osmo_io_uring, ring);
+	struct iofd_msghdr *msghdr;
 
 	while (io_uring_peek_cqe(ring, &cqe) == 0) {
-		orig_ring->num_pending_submissions--;
 
 		msghdr = io_uring_cqe_get_data(cqe);
 		if (!msghdr) {
@@ -458,8 +274,6 @@ static void iofd_uring_cqe(struct io_uring *ring)
 		}
 		if (!msghdr->iofd) {
 			io_uring_cqe_seen(ring, cqe);
-			if (msghdr->in_cancel_queue)
-				llist_del(&msghdr->list);
 			iofd_msghdr_free(msghdr);
 			continue;
 		}
@@ -471,26 +285,6 @@ static void iofd_uring_cqe(struct io_uring *ring)
 		iofd_uring_handle_completion(msghdr, rc);
 
 	}
-
-	/* If there are unsubmitted cancel SQEs, try to queue them now. */
-	if (OSMO_LIKELY(llist_empty(&orig_ring->cancel_queue))) {
-		/* Old ring is empty, remove it. */
-		if (OSMO_UNLIKELY(orig_ring != g_ring && orig_ring->num_pending_submissions == 0))
-			osmo_iofd_uring_exit(orig_ring);
-		return;
-	}
-	llist_for_each_entry_safe(msghdr, msghdr2, &orig_ring->cancel_queue, list) {
-		struct io_uring_sqe *sqe;
-		sqe = io_uring_get_sqe_and_count(orig_ring);
-		if (!sqe)
-			break;
-		io_uring_sqe_set_data(sqe, NULL);
-		LOGP(DLIO, LOGL_DEBUG, "Cancelling queued read/write\n");
-		io_uring_prep_cancel(sqe, msghdr, 0);
-		llist_del(&msghdr->list);
-		msghdr->in_cancel_queue = false;
-	}
-	io_uring_submit(&orig_ring->ring);
 }
 
 /*! will submit the next to-be-transmitted message for given iofd */
@@ -503,7 +297,7 @@ static int iofd_uring_submit_tx(struct osmo_io_fd *iofd)
 	if (!msghdr)
 		return -ENODATA;
 
-	sqe = iofd_uring_get_sqe(iofd, false);
+	sqe = io_uring_get_sqe(&g_ring.ring);
 	if (!sqe) {
 		LOGPIO(iofd, LOGL_ERROR, "Could not get io_uring_sqe\n");
 		OSMO_ASSERT(0);
@@ -513,7 +307,7 @@ static int iofd_uring_submit_tx(struct osmo_io_fd *iofd)
 
 	switch (msghdr->action) {
 	case IOFD_ACT_WRITE:
-		io_uring_prep_writev(sqe, msghdr->iofd->fd, msghdr->iov, msghdr->io_len, -1);
+		io_uring_prep_writev(sqe, msghdr->iofd->fd, msghdr->iov, 1, -1);
 		break;
 	case IOFD_ACT_SENDTO:
 	case IOFD_ACT_SENDMSG:
@@ -523,10 +317,8 @@ static int iofd_uring_submit_tx(struct osmo_io_fd *iofd)
 		OSMO_ASSERT(0);
 	}
 
-	iofd_io_uring_submit();
-
-	iofd->u.uring.write.msghdr = msghdr;
-	iofd->u.uring.write.ring = &g_ring->ring;
+	io_uring_submit(&g_ring.ring);
+	iofd->u.uring.write_msghdr = msghdr;
 
 	return 0;
 }
@@ -565,23 +357,16 @@ static int iofd_uring_connected_cb(struct osmo_fd *ofd, unsigned int what)
 	IOFD_FLAG_UNSET(iofd, IOFD_FLAG_IN_CALLBACK);
 
 	/* If write/read notifications are pending, enable it now. */
-	if (iofd->u.uring.write.enabled && !IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))
+	if (iofd->u.uring.write_enabled && !IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))
 		iofd_uring_write_enable(iofd);
-	if (iofd->u.uring.read.enabled && !IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))
+	if (iofd->u.uring.read_enabled && !IOFD_FLAG_ISSET(iofd, IOFD_FLAG_CLOSED))
 		iofd_uring_read_enable(iofd);
 
-	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_TO_FREE) && !iofd->u.uring.read.sqes_submitted && !iofd->u.uring.write.msghdr)
+	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_TO_FREE) && !iofd->u.uring.read_msghdr && !iofd->u.uring.write_msghdr)
 		talloc_free(iofd);
 	return 0;
 }
 
-static int iofd_uring_setup(struct osmo_io_fd *iofd)
-{
-	iofd->u.uring.read.num_sqes = g_io_uring_read_sqes;
-
-	return 0;
-}
-
 static int iofd_uring_register(struct osmo_io_fd *iofd)
 {
 	if (iofd->mode != OSMO_IO_FD_MODE_RECVMSG_SENDMSG)
@@ -607,60 +392,32 @@ static int iofd_uring_unregister(struct osmo_io_fd *iofd)
 {
 	struct io_uring_sqe *sqe;
 	struct iofd_msghdr *msghdr;
-	uint8_t idx;
 
-	for (idx = 0; idx < iofd->u.uring.read.sqes_submitted; idx++) {
-		struct osmo_io_uring *ring = container_of(iofd->u.uring.read.ring, struct osmo_io_uring, ring);
-		msghdr = iofd->u.uring.read.msghdr[idx];
-		iofd->u.uring.read.msghdr[idx] = NULL;
-		/* Submit SQEs of the current ring, if needed. */
-		if (&ring->ring == &g_ring->ring)
-			osmo_io_uring_submit();
-		/* If the submission queue is full, use cancel queue. We cannot cancel SQEs on the new ring. */
-		sqe = io_uring_get_sqe_and_count(ring);
-		if (sqe) {
-			io_uring_sqe_set_data(sqe, NULL);
-			LOGPIO(iofd, LOGL_DEBUG, "Cancelling read\n");
-			io_uring_prep_cancel(sqe, msghdr, 0);
-			io_uring_submit(&ring->ring);
-		} else {
-			llist_add_tail(&msghdr->list, &ring->cancel_queue);
-			msghdr->in_cancel_queue = true;
-		}
+	if (iofd->u.uring.read_msghdr) {
+		msghdr = iofd->u.uring.read_msghdr;
+		sqe = io_uring_get_sqe(&g_ring.ring);
+		OSMO_ASSERT(sqe != NULL);
+		io_uring_sqe_set_data(sqe, NULL);
+		LOGPIO(iofd, LOGL_DEBUG, "Cancelling read\n");
+		iofd->u.uring.read_msghdr = NULL;
 		talloc_steal(OTC_GLOBAL, msghdr);
 		msghdr->iofd = NULL;
-	}
-	if (iofd->u.uring.read.sqes_submitted) {
-		iofd->u.uring.read.ring = NULL;
-		iofd->u.uring.read.sqes_submitted = 0;
+		io_uring_prep_cancel(sqe, msghdr, 0);
 	}
 
-	if (iofd->u.uring.write.msghdr) {
-		struct osmo_io_uring *ring = container_of(iofd->u.uring.write.ring, struct osmo_io_uring, ring);
-		msghdr = iofd->u.uring.write.msghdr;
-		iofd->u.uring.write.msghdr = NULL;
-		for (int idx = 0; idx < msghdr->io_len; idx++) {
-			msgb_free(msghdr->msg[idx]);
-			msghdr->msg[idx] = NULL;
-		}
-		/* Submit SQEs of the current ring, if needed. */
-		if (&ring->ring == &g_ring->ring)
-			osmo_io_uring_submit();
-		/* If the submission queue is full, use cancel queue. We cannot cancel SQEs on the new ring. */
-		sqe = io_uring_get_sqe_and_count(ring);
-		if (sqe) {
-			io_uring_sqe_set_data(sqe, NULL);
-			LOGPIO(iofd, LOGL_DEBUG, "Cancelling write\n");
-			io_uring_prep_cancel(sqe, msghdr, 0);
-			io_uring_submit(&ring->ring);
-		} else {
-			llist_add_tail(&msghdr->list, &ring->cancel_queue);
-			msghdr->in_cancel_queue = true;
-		}
+	if (iofd->u.uring.write_msghdr) {
+		msghdr = iofd->u.uring.write_msghdr;
+		sqe = io_uring_get_sqe(&g_ring.ring);
+		OSMO_ASSERT(sqe != NULL);
+		io_uring_sqe_set_data(sqe, NULL);
+		LOGPIO(iofd, LOGL_DEBUG, "Cancelling write\n");
+		iofd->u.uring.write_msghdr = NULL;
 		talloc_steal(OTC_GLOBAL, msghdr);
+		msgb_free(msghdr->msg);
 		msghdr->iofd = NULL;
-		iofd->u.uring.write.ring = NULL;
+		io_uring_prep_cancel(sqe, msghdr, 0);
 	}
+	io_uring_submit(&g_ring.ring);
 
 	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_NOTIFY_CONNECTED)) {
 		osmo_fd_unregister(&iofd->u.uring.connect_ofd);
@@ -672,9 +429,9 @@ static int iofd_uring_unregister(struct osmo_io_fd *iofd)
 
 static void iofd_uring_write_enable(struct osmo_io_fd *iofd)
 {
-	iofd->u.uring.write.enabled = true;
+	iofd->u.uring.write_enabled = true;
 
-	if (iofd->u.uring.write.msghdr)
+	if (iofd->u.uring.write_msghdr)
 		return;
 
 	/* This function is called again, once the socket is connected. */
@@ -701,7 +458,7 @@ static void iofd_uring_write_enable(struct osmo_io_fd *iofd)
 		msghdr->iov[0].iov_base = msgb_data(msg);
 		msghdr->iov[0].iov_len = msgb_length(msg);
 
-		sqe = iofd_uring_get_sqe(iofd, false);
+		sqe = io_uring_get_sqe(&g_ring.ring);
 		if (!sqe) {
 			LOGPIO(iofd, LOGL_ERROR, "Could not get io_uring_sqe\n");
 			OSMO_ASSERT(0);
@@ -709,23 +466,21 @@ static void iofd_uring_write_enable(struct osmo_io_fd *iofd)
 		io_uring_prep_writev(sqe, iofd->fd, msghdr->iov, 1, 0);
 		io_uring_sqe_set_data(sqe, msghdr);
 
-		iofd_io_uring_submit();
-
-		iofd->u.uring.write.msghdr = msghdr;
-		iofd->u.uring.write.ring = &g_ring->ring;
+		io_uring_submit(&g_ring.ring);
+		iofd->u.uring.write_msghdr = msghdr;
 	}
 }
 
 static void iofd_uring_write_disable(struct osmo_io_fd *iofd)
 {
-	iofd->u.uring.write.enabled = false;
+	iofd->u.uring.write_enabled = false;
 }
 
 static void iofd_uring_read_enable(struct osmo_io_fd *iofd)
 {
-	iofd->u.uring.read.enabled = true;
+	iofd->u.uring.read_enabled = true;
 
-	if (iofd->u.uring.read.sqes_submitted)
+	if (iofd->u.uring.read_msghdr)
 		return;
 
 	/* This function is called again, once the socket is connected. */
@@ -749,7 +504,7 @@ static void iofd_uring_read_enable(struct osmo_io_fd *iofd)
 
 static void iofd_uring_read_disable(struct osmo_io_fd *iofd)
 {
-	iofd->u.uring.read.enabled = false;
+	iofd->u.uring.read_enabled = false;
 }
 
 static int iofd_uring_close(struct osmo_io_fd *iofd)
@@ -765,14 +520,13 @@ static void iofd_uring_notify_connected(struct osmo_io_fd *iofd)
 	}
 
 	/* OSMO_IO_FD_MODE_RECVMSG_SENDMSG: Don't call this function after enabling read or write. */
-	OSMO_ASSERT(!iofd->u.uring.write.enabled && !iofd->u.uring.read.enabled);
+	OSMO_ASSERT(!iofd->u.uring.write_enabled && !iofd->u.uring.read_enabled);
 
 	/* Set flag to enable temporary osmo_fd during register() time: */
 	IOFD_FLAG_SET(iofd, IOFD_FLAG_NOTIFY_CONNECTED);
 }
 
 const struct iofd_backend_ops iofd_uring_ops = {
-	.setup = iofd_uring_setup,
 	.register_fd = iofd_uring_register,
 	.unregister_fd = iofd_uring_unregister,
 	.close = iofd_uring_close,
@@ -783,12 +537,4 @@ const struct iofd_backend_ops iofd_uring_ops = {
 	.notify_connected = iofd_uring_notify_connected,
 };
 
-void osmo_io_uring_submit(void)
-{
-	if (OSMO_LIKELY(g_io_uring_submit_needed)) {
-		io_uring_submit(&g_ring->ring);
-		g_io_uring_submit_needed = false;
-	}
-}
-
 #endif /* defined(__linux__) */
